

#include <stdint.h>


using namespace std;


extern "C" {
	extern void (*__CTOR_LIST__)();
	extern void (*__DTOR_LIST__)();
	extern void (*__init_array_start)();
	extern void (*__init_array_end)();
	extern void (*__fini_array_start)();
	extern void (*__fini_array_end)();
}


void _callInitArray() {
	void (**f)() = &__init_array_start;
	while (f != &__init_array_end) {
		(*f)();
		f++;
	}
}


void _callConstructors() {
	void (**constructor)() = &__CTOR_LIST__;
	uint32_t total = *(uint32_t *) constructor;
	constructor++;
	while (total) {
		(*constructor)();
		total--;
		constructor++;
	}
}


void _callDestructors() {
	void (**destructor)() = &__DTOR_LIST__;
	uint32_t total = *(uint32_t *) destructor;
	destructor++;
	while (total) {
		(*destructor)();
		total--;
		destructor++;
	}
}


void _callFiniArray() {
	void (**f)() = &__fini_array_start;
	while (f != &__fini_array_end) {
		(*f)();
		f++;
	}
}


#define  SYSPLLCTRL    *((uint32_t *) 0x40048008)
#define  PDRUNCFG      *((uint32_t *) 0x40048238)
#define  SYSPLLSTAT    *((uint32_t *) 0x4004800C)
#define  MAINCLKSEL    *((uint32_t *) 0x40048070)
#define  MAINCLKUEN    *((uint32_t *) 0x40048074)
#define  SYSAHBCLKDIV  *((uint32_t *) 0x40048078)


void _initClock() {
	// configure PLL: 12 MHz --> 60 MHz
	// http://blog.nano-age.co.uk/2014/07/the-pll-and-various-clock-sources-of.html
	SYSPLLCTRL = 0x24;
	 // power up PLL
	PDRUNCFG &= ~(((uint32_t) 1) << 7);
	// wait for PLL to be locked
	while (!(SYSPLLSTAT & 1))
		;
	// main clock --> pll output
	MAINCLKSEL = 3;
	// update main clock source
	MAINCLKUEN = 0;
	MAINCLKUEN = 1;
	// system clock = 60 MHz / 2 = 30 MHz
	SYSAHBCLKDIV = 2;
}


extern int main();


extern "C" {
	extern unsigned char flash_sdata;
	extern unsigned char ram_sdata;
	extern unsigned char ram_edata;
}


void _initDataRAM() {
	// init .data section (global variables) with flash data
	unsigned char *from = &flash_sdata;
	unsigned char *to = &ram_sdata;
	while (to != &ram_edata) {
		*to = *from;
		from++;
		to++;
	}
}


extern "C" {
	extern unsigned char ram_sbssdata;
	extern unsigned char ram_ebssdata;
}


void _initBssRAM() {
	// init .bss section with zeros
	unsigned char *p = &ram_sbssdata;
	while (p != &ram_ebssdata) {
		*p = 0;
		p++;
	}
}


void _startup() __attribute__((section(".startup"), naked));


void _startup() {
	_initDataRAM();
	_initBssRAM();
	_initClock();
	_callConstructors();
	_callInitArray();
	main();
	_callFiniArray();
	_callDestructors();
	while (true)
		;
}

extern "C" void __cxa_pure_virtual() {}
void *__dso_handle = 0;
extern "C" void __cxa_atexit() {}

